import hashlib
import json
from dataclasses import dataclass
from enum import Enum

type JsonValue = (
    list[JsonValue] | dict[str, JsonValue] | str | bool | int | float | None
)


def fully_qualified_name(tp: type) -> str:
    mod = tp.__module__
    qualname = tp.__qualname__
    if mod == "__main__":  # TODO: allow overwriting
        raise ValueError("Cannot serialize objects from __main__ module")
    elif "<locals>" in mod:  # TODO: allow overwriting
        raise ValueError("TODO: msg")
    elif "." in qualname:
        raise ValueError("TODO: msg")
    elif (isinstance(tp, type) and issubclass(tp, Enum)) or isinstance(tp, Enum):
        raise ValueError(
            "TODO: support this in the future"
        )  # return f"{mod}.{qualname}.{obj.name}"
    return f"{mod}.{qualname}"


def _stable_json_dump(x: JsonValue) -> str:
    return json.dumps(x, sort_keys=True, separators=(",", ":"))


def _hash_dict_deterministically(obj: JsonValue) -> str:
    json_str = _stable_json_dump(obj)

    return hashlib.blake2s(
        json_str.encode(),
        digest_size=10,  # TODO: make this digest size configurable and include a script for estimating likelihood of crashing. right now, i think there is a 1e-08 chance of a collision with 155M items with the same schema and namespace
    ).hexdigest()


@dataclass(frozen=True, slots=True, kw_only=True)
class Ok[T]:
    result: T



import enum
from dataclasses import fields, is_dataclass
from pathlib import Path
from typing import Any

from pydantic import BaseModel as PydanticBaseModel

from furu.constants import CLASSMARKER
from furu.utils import JsonValue, fully_qualified_name


def to_json(  # TODO: consider caching this (but if i'm going to, I need to figure out how to cache lists and other unhashable objects)
    obj: Any,
) -> JsonValue:
    # TODO: when writing this to metadata, make sure to escape strings etc

    def assert_correct_dict_key(x: Any) -> str:
        if not isinstance(x, str):
            raise ValueError("TODO")
        if x == CLASSMARKER:
            raise ValueError("TODO: write error msg")
        return x

    match obj:
        case None:
            return None
        case int() | str() | float() | bool():
            return obj
        case Path():
            return str(obj)
        case list() | tuple():
            return [to_json(x) for x in obj]
        case set() | frozenset():
            return sorted([to_json(x) for x in obj], key=repr)
        case dict():
            return {assert_correct_dict_key(k): to_json(v) for k, v in obj.items()}
        case x if is_dataclass(x):
            return {
                CLASSMARKER: fully_qualified_name(type(x)),
                **{f.name: to_json(getattr(x, f.name)) for f in fields(x)},
            }
        case PydanticBaseModel():
            return {
                CLASSMARKER: fully_qualified_name(type(obj)),
                **{k: to_json(v) for k, v in obj.model_dump().items()},
            }
        case enum.Enum():
            raise NotImplementedError("TODO")  #  return {"__enum__": _type_fqn(obj)}
        case _:
            raise ValueError("unexpected item", obj)  # TODO: explain the error more



from typing import Literal

from pydantic import BaseModel, ConfigDict


class State(BaseModel):
    model_config = ConfigDict(extra="forbid", validate_assignment=True, strict=True)


class SuccessState(State):
    status: Literal["success"] = "success"


class FailedState(State):
    status: Literal["failed"] = "failed"


class RunningState(State):
    status: Literal["running"] = "running"



import enum
import types
import typing
from dataclasses import fields, is_dataclass
from typing import (
    Any,
    get_args,
    get_origin,
    get_type_hints,
)

from furu.constants import ARGSMARKER, CLASSMARKER, ORIGINMARKER
from furu.utils import JsonValue, _stable_json_dump, fully_qualified_name


def schema_dataclass(tp: type, seen: set[type]) -> JsonValue:
    if tp in seen:
        return {CLASSMARKER: fully_qualified_name(tp)}
    seen.add(tp)

    hints = get_type_hints(tp, include_extras=True)
    return {
        CLASSMARKER: fully_qualified_name(tp),
        "fields": {
            f.name: schema_type(hints[f.name], seen)
            for f in sorted(fields(tp), key=lambda f: f.name)
            if not f.name.startswith("_")
        },
    }


def schema_type(tp: Any, seen: set[type]) -> JsonValue:
    origin = get_origin(tp)

    if tp is Ellipsis:
        return fully_qualified_name(types.EllipsisType)

    if isinstance(tp, type) and is_dataclass(tp):
        return schema_dataclass(tp, seen)
    if origin is not None and is_dataclass(origin):
        return schema_dataclass(origin, seen)

    if origin in (typing.Union, types.UnionType):
        return sorted(
            [schema_type(a, seen) for a in get_args(tp)], key=_stable_json_dump
        )
    elif origin is not None:
        assert (args := get_args(tp))  # TODO: maybe i need to remove this?
        return {
            ORIGINMARKER: fully_qualified_name(origin),
            ARGSMARKER: sorted(
                [schema_type(a, seen) for a in args], key=_stable_json_dump
            ),
        }

    if tp in [str, float, int, bool, types.NoneType]:
        return fully_qualified_name(tp)
    elif isinstance(tp, str):
        return tp
    elif tp in [list, tuple, dict]:
        if get_args(tp) != ():
            raise ValueError(f"Expected bare {tp.__name__}, got parameterized type")
        return fully_qualified_name(tp)
    elif isinstance(tp, typing.TypeVar):
        return repr(tp)
    elif isinstance(tp, enum.EnumType):
        return fully_qualified_name(tp)
    elif isinstance(tp, type):
        return fully_qualified_name(tp)
    assert False, f"TODO: unexpected type value {tp=}"



from datetime import datetime
from pathlib import Path

from pydantic import BaseModel, ConfigDict

from furu.utils import JsonValue


class GitData(BaseModel):
    commit: str
    branch: str
    remote: str
    # patch: dict # TODO: add this
    # submodules: dict # TODO: add this


class _Metadata(BaseModel):
    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        strict=True,
        revalidate_instances="always",
    )
    # python_def: str
    artifact: JsonValue
    artifact_hash: str
    schema_: JsonValue
    schema_hash: str
    data_path: Path
    # git: GitData | None


class RunningMetadata(_Metadata):
    started_at: datetime  # TODO: consider making this timezone aware
    # command: list[str] # TODO: find/decide what the most elegant approach is here
    # hostname: str
    # user: str
    # pid: int
    # TODO: what other information should i record about the person starting this run?
    # furu_package_version: str
    # TODO: include some hardware information and machine info, such as os/version
    # python_version: str
    # executor info, such as local, local executor, slurm dag or slurm worker


class CompletedMetadata(RunningMetadata):
    # traced_function_hashes: list[
    #     dict[str, str]
    # ]  # TODO: this probably means i don't really need to record the package versions? in particular since git data would have uv.lock and pyproject.toml already
    # slurm: SlurmData
    completed_at: datetime



import pickle
import secrets
import traceback
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from functools import cache, cached_property
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    Self,
    assert_never,
)

from furu.config import config
from furu.locking import run_with_lease_and_pickle_result
from furu.metadata import CompletedMetadata, RunningMetadata
from furu.schema import schema_type as _schema_type
from furu.serialize import to_json as _to_json
from furu.utils import JsonValue, Ok, _hash_dict_deterministically, fully_qualified_name

if TYPE_CHECKING:
    from typing_extensions import dataclass_transform

    @dataclass_transform(kw_only_default=True, frozen_default=True)
    class _FuruDataclassTransform:
        pass
else:

    class _FuruDataclassTransform:
        pass


class Furu[T](_FuruDataclassTransform, ABC):
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if cls is Furu:
            return
        if "__dataclass_params__" not in cls.__dict__:
            dataclass(frozen=True, kw_only=True)(cls)

    def load_or_create(self) -> T:
        if self._result_path.exists():
            # TODO: validation that its up to date and valid
            with open(self._result_path, "rb") as f:
                return pickle.load(f)

        # TODO: have the logs be per-run with a run id

        # TODO: initialize the state
        # TODO: add file locking and keepalive in a different process
        self._internal_furu_dir.mkdir(
            exist_ok=True, parents=True
        )  # TODO: decide if i should make the directory here or inside the cached property itself

        def create_wrapper() -> T:  # TODO: better name
            metadata = RunningMetadata(
                artifact=self.to_json(),
                artifact_hash=self.artifact_hash,
                schema_=self.schema,
                schema_hash=self.schema_hash,
                data_path=self.data_dir.resolve(),
                started_at=datetime.now(),  # TODO: make this timezone aware?
            )
            self._metadata_path.write_text(metadata.model_dump_json(indent=2))

            try:
                result = self._create()
            except BaseException as exc:
                with (  # TODO: log this to the regular log file
                    (
                        self._internal_furu_dir
                        / f"error-{datetime.now():%y%m%d_%H-%M-%S}-{secrets.token_hex(4)}.log"  # TODO: make this part of the regular error
                    ).open("a", encoding="utf-8") as f
                ):
                    f.write("Traceback (most recent call last):\n")
                    f.writelines(
                        traceback.format_list(
                            traceback.extract_stack()[:-1]
                            + traceback.extract_tb(exc.__traceback__)
                        )
                    )
                    f.writelines(traceback.format_exception_only(type(exc), exc))
                    f.write("\n=== Debug Details (with locals) ===\n")
                    f.writelines(
                        traceback.TracebackException.from_exception(
                            exc, capture_locals=True
                        ).format(chain=True)
                    )
                raise

            completed_metadata = CompletedMetadata(
                artifact=metadata.artifact,
                artifact_hash=metadata.artifact_hash,
                schema_=metadata.schema_,
                schema_hash=metadata.schema_hash,
                data_path=metadata.data_path,
                started_at=metadata.started_at,
                completed_at=datetime.now(),
            )
            self._metadata_path.write_text(completed_metadata.model_dump_json(indent=2))

            return result

        maybe_result = run_with_lease_and_pickle_result(
            create_wrapper,
            lock_path=self._internal_furu_dir / "compute.lock",
            result_path=self._result_path,
        )

        match maybe_result:
            case Ok(result=result):
                return result
            case "lost-lock":
                raise NotImplementedError("TODO: Handle this")
            case "missing-tmp":
                raise NotImplementedError("TODO: Handle this")
            case x:
                assert_never(x)

    def status(
        self,
    ) -> Literal[
        "completed", "missing", "running", "failed"
    ]:  # TODO: add queued/waiting state?
        if self.is_completed():
            return "completed"
        raise NotImplementedError("TODO")

    def is_completed(
        self,
    ) -> bool:  # TODO: maybe this should check the is self.status is completed? (in that case status cant check if self.is_completed)
        return self._result_path.exists()

    def try_load(self) -> T:  # TODO: make a better name for this
        if self._result_path.exists():
            # TODO: validation that its up to date and valid
            with open(self._result_path, "rb") as f:
                return pickle.load(f)
        raise NotImplementedError(
            "TODO: decide if i should throw or return error value"
        )

    @property
    def _result_path(self) -> Path:
        return self.data_dir / "result.pkl"

    @abstractmethod
    def _create(self) -> T:
        raise NotImplementedError("TODO")

    @cache
    def to_json(  # TODO: make sure this doesn't prevent garbage collection
        self,
    ) -> JsonValue:
        return _to_json(self)

    @classmethod
    def from_json(cls) -> Self:
        raise NotImplementedError("TODO")

    @cached_property
    def artifact_hash(  # TODO: should this be __hash__?
        self,
    ) -> str:
        return _hash_dict_deterministically(self.to_json())

    @cached_property
    def schema(
        self,
    ) -> JsonValue:
        return _schema_type(type(self), set())

    @cached_property
    def schema_hash(self) -> str:
        return _hash_dict_deterministically(self.schema)

    @cached_property  # TODO: decide if something like this should be cached_property or simply property
    def data_dir(self) -> Path:
        return (
            config.directories.data
            / Path(*fully_qualified_name(type(self)).split("."))
            / self.schema_hash
            / self.artifact_hash
        )

    @cached_property
    def _internal_furu_dir(self) -> Path:
        return self.data_dir / ".furu"

    @cached_property
    def _metadata_path(self) -> Path:
        return self._internal_furu_dir / "metadata.json"



# TODO: this is the first iteration of this file, but it has multiple errors and footguns, such as the gil being busy for more than LeaseConfig.lifetime_s or the heartbeat thread losing the lock but not notifying/killing the worker. i will either move this to zig or rewrite it at some point
# TODO: handle cases where a user submits a 8 task torch job
import os
import pickle
import threading
from collections.abc import Callable
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Literal

from flufl.lock import Lock, NotLockedError, TimeOutError

from furu.utils import Ok

type LeaseErrorStates = Literal["lost-lock", "missing-tmp"]


@dataclass(frozen=True, slots=True)
class LeaseConfig:
    lease_ttl_s: int = 120
    review_interval_s: int = 15


def _compute_and_stage_pickle_result[T](
    *,
    staged_result_path: Path,
    compute_fn: Callable[[], T],
) -> T:
    result = compute_fn()
    with staged_result_path.open("wb") as f:
        pickle.dump(
            result,
            f,
        )
        f.flush()  # TODO: Do i need this and the os.fsync?
        os.fsync(f.fileno())
    return result


def run_with_lease_and_pickle_result[T](
    # data_dir: Path,
    compute_fn: Callable[[], T],
    *,
    lock_path: Path,
    result_path: Path,
    lease_config: LeaseConfig | None = None,
    # TODO: add a n-retries option
) -> Ok | LeaseErrorStates:
    if lease_config is None:
        lease_config = LeaseConfig()
    # TODO: make the directory for the data and the <data-dir>/.furu
    # TODO: if the gil is busy for more than lease_config.lifetime_s we will lose the lock. move to zig or use multiprocessing. the reason we are not already using multiprocessing is that it is more painful to return the result from the self._create call when we use multiprocessing
    stop_lease_heartbeat = threading.Event()
    lease_lost = threading.Event()
    now = datetime.now()
    staged_result_path = result_path.with_suffix(
        f".{os.getpid()}-{now:%y%m%d_%H-%M-%S}.pkl.tmp"
    )  # TODO: compute this correctly

    try:
        file_lock = Lock(str(lock_path), lifetime=lease_config.lease_ttl_s)

        try:
            file_lock.lock(timeout=0)  # TODO:make sure timeout 0 makes sense
        except TimeOutError as e:
            print(f"failed to lock {file_lock} at path {str(lock_path)}")
            raise NotImplementedError(
                "TODO: wait for the object to finish and return/read a copy for some time, die if you wait for too long and check if the old lease is expired (should be automatic)"
            ) from e

        def review_lease_loop() -> None:
            while not stop_lease_heartbeat.wait(lease_config.review_interval_s):
                try:
                    file_lock.refresh()
                except NotLockedError:
                    lease_lost.set()
                    stop_lease_heartbeat.set()
                    return  # TODO: How do i fail/throw correctly here?

        lease_heartbeat_thread = threading.Thread(target=review_lease_loop, daemon=True)
        lease_heartbeat_thread.start()

        try:
            result = _compute_and_stage_pickle_result(
                staged_result_path=staged_result_path, compute_fn=compute_fn
            )
        except BaseException:
            staged_result_path.unlink(missing_ok=True)
            # TODO: maybe handle BaseException and Exception differently?
            raise

        if not staged_result_path.exists():
            return "missing-tmp"
        elif lease_lost.is_set():
            staged_result_path.unlink()
            return "lost-lock"

        try:
            file_lock.refresh()
        except NotLockedError:
            staged_result_path.unlink()
            return "lost-lock"

        os.replace(staged_result_path, result_path)

        return Ok(result=result)

    finally:
        stop_lease_heartbeat.set()

        # TODO: this doesn't actually kill the heartbeat thread

        if staged_result_path.exists():
            raise NotImplementedError("TODO: i don't think this can ever happen?")

        file_lock.unlock()  # TODO: this can technically fail if we didn't aquire the lock. won't fix this since this logic will be moved to zig



from dataclasses import dataclass, field
from pathlib import Path
from typing import Self


@dataclass(slots=True)
class _FuruDirectories:
    data: Path
    # TODO: make this better and more user configurable, so that it is easy for the user to define exactly which paths they want to save to and which furu objects should save where

    @classmethod
    def default(cls) -> Self:
        # TODO: make sure this location is deterministic/more predictable, such as by finding the next .git or pyproject.toml or furu directory
        base_dir = Path("furu-data")
        data_dir = base_dir / "data"
        return cls(data=data_dir)


@dataclass(slots=True)
class _FuruConfig:
    debug_mode: bool = False
    directories: _FuruDirectories = field(default_factory=_FuruDirectories.default)
    # base_directory: Path = Path(
    #     "furu-data"
    # )  # TODO: allow this to be multiple different paths too


config = _FuruConfig()



import hashlib
import os
import secrets
import shutil
import tempfile
from collections.abc import Iterator
from dataclasses import dataclass, fields
from datetime import datetime
from pathlib import Path

import pytest

from furu.config import _FuruDirectories
from furu.config import config as furu_config


@dataclass(slots=True)
class _FuruPytestState:  # TODO: maybe make this auto snapshot the config
    original_directories: _FuruDirectories
    run_directories: _FuruDirectories


_STATE_KEY = pytest.StashKey[_FuruPytestState]()


def _is_furu_pytest_mode_enabled() -> bool:
    return os.environ.get("FURU_PYTEST_MODE", "on").strip().lower() != "off"


def _keep_furu_data() -> bool:
    return os.environ.get("FURU_PYTEST_KEEP", "clean").strip().lower() == "keep"


def pytest_configure(config: pytest.Config) -> None:
    if not _is_furu_pytest_mode_enabled():
        return

    run_base_directory = (
        Path(tempfile.gettempdir())
        / f"furu-data-{datetime.now().strftime('%Y%m%d-%H%M%S')}-{secrets.token_hex(4)}"
    )
    # run_base_directory.mkdir(parents=True, exist_ok=True)

    state = _FuruPytestState(
        original_directories=furu_config.directories,
        run_directories=_FuruDirectories(
            data=run_base_directory,
        ),
    )

    furu_config.directories = state.run_directories
    config.stash[_STATE_KEY] = state


def pytest_unconfigure(config: pytest.Config) -> None:
    if not _is_furu_pytest_mode_enabled():
        return

    state = config.stash[_STATE_KEY]

    furu_config.directories = state.original_directories

    if not _keep_furu_data():
        for field in fields(state.run_directories):
            path = getattr(state.run_directories, field.name)
            if not isinstance(path, Path):
                raise TypeError(
                    f"Expected Path for run directory '{field.name}', got {type(path).__name__}"
                )
            shutil.rmtree(path, ignore_errors=True)
    else:
        print(f"kept furu data at {state.run_directories.data}")


@pytest.fixture(autouse=True)
def _furu_per_test_base_directory(
    request: pytest.FixtureRequest,
    pytestconfig: pytest.Config,
) -> Iterator[None]:
    if not _is_furu_pytest_mode_enabled():
        yield
        return

    state = pytestconfig.stash[_STATE_KEY]
    previous_base_directory = furu_config.directories

    test_base_directory = (
        state.run_directories.data
        / hashlib.sha1(request.node.nodeid.encode("utf-8")).hexdigest()[:12]
    )

    furu_config.directories = _FuruDirectories(data=test_base_directory)
    try:
        yield
    finally:
        furu_config.directories = previous_base_directory
